/*
 * IBM Confidential
 *
 * OCO Source Materials
 *
 * IBM XSLT Processor (XSLT4J)
 *
 * (C) Copyright IBM Corp. 2007, 2008
 *
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
 */
package org.apache.xml.utils;

import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;

/**
 * Utility class for defining hidden options that can be used to enable
 * debugging trace information or experimental features.  Options are set
 * using the <code>com.ibm.xslt4j.options</code> system property.  Option
 * syntax is as follows:
 * <p><code>
 * options  ::=  option [ ";" option ]*<br>  
 * option  ::=  name [ value ]<br> 
 * value  ::=  "(" options ")"<br>
 * name ::= ( char - [:()] )+</code>
 * <p>Options have optional values that are enclosed in parentheses.  Options
 * are separated by colons.  Option values themselves are colon separated
 * options.  A top-level option has itself as its name; the name of a suboption
 * is the name of its parent option concatenated with a period followed by the
 * suboption. 
  * <p>
 * For example, consider the following system property, set on the JVM:
 * <code>
 *   -Dcom.ibm.xslt4j.options=a:b(c(true):d(e):f):g(10)
 * </code>
 * This specifies the following options:
 * <ul>
 * <li> "a"
 * <li> "b"
 * <li> "b.c"
 * <li> "b.c.true"
 * <li> "b.d"
 * <li> "b.d.e"
 * <li> "b.f"
 * <li> "g"
 * <li> "g.10"
 * </ul>
 * When any of thos string keys are passed to wasSpecified() is will return true,
 * otherwise that method returns false.
 * <p>
 * When those string keys are passed to getStringValue(), 
 * it would return the following values respectively:
 * <ul>
 * <li> null
 * <li> "c(true):d(e):f"
 * <li> "true"
 * <li> null
 * <li> "e"
 * <li> null
 * <li> null
 * <li> "10"
 * <li> null
 * </ul>
 * <p>
 * When those same keys are passed to getIntegerValue(), 
 * it would return the following values respectively:
 * <ul>
 * <li> null
 * <li> null
 * <li> null
 * <li> null
 * <li> null
 * <li> null
 * <li> null
 * <li> an Integer object, which has an intValue() method that returns 10.
 * <li> null
 * </ul>
 * 
 * <p>
 * When those same keys are passed to getBooleanValue(), 
 * it would return the following values respectively:
 * <ul>
 * <li> null
 * <li> null
 * <li> a java.lang.Boolean object, which has booleanValue() method that returns true.
 * <li> null
 * <li> null
 * <li> null
 * <li> null
 * <li> null
 * <li> null
 * </ul>
 * <p>
 * If you want to trace what is happening with the hidden options, indeed to even watch
 * which ones are being queried you should specify the "HiddenOptions", for example set this the JVM:
 * <code>
 *   -Dcom.ibm.xslt4j.options=HiddenOptions:a:b(c(true):d(e):f):g(10)
 * </code>
 *  
 */
final public class HiddenOptions {
    /**
     * Hidden constructor.
     */
    private HiddenOptions() {
    }

    /**
     * Contains debug options and their values.
     */
    final private static HashMap sOptions;

    /**
     * Has debugging of this class been enabled? 
     */
    final private static boolean sDebugMe;

    final private static String DEBUG_PROPERTY = "com.ibm.xslt4j.options";
    static {
        // Make sure to access the system property using the authority of this
        // class rather than that of any user code on the stack when this class
        // was first referenced.
        final String debugOptions =
            (String) AccessController.doPrivileged(new PrivilegedAction() {
                         public Object run() {
                             return System.getProperty(DEBUG_PROPERTY);
                         }
                     });

        boolean forceDebuggingForMe = false;

        // Examine system property setting 
        if (debugOptions != null) {
            sOptions = new HashMap();
            try {
                // installOptions(sOptions, debugOptions, "", 0);
                installOptions(sOptions, debugOptions);
            } catch (Throwable t) {
                // Ignore any exceptions - options string might have been
                // incompletely processed.
                forceDebuggingForMe = true;
            }

            // Check for self-debugging
            sDebugMe = forceDebuggingForMe
                           || wasSpecified("HiddenOptions");

            if (sDebugMe) {
                System.err.println(HiddenOptions.class.getName()
                        + ":  value of " + DEBUG_PROPERTY + " == "
                        + debugOptions);
                System.err.println(HiddenOptions.class.getName() + ":  options "
                                       + dump());
            }
        } else {
            sOptions = null;
            sDebugMe = forceDebuggingForMe;
        }
    }

    /**
     * Parse debug options and set options map accordingly.  See
     * {@link HiddenOptions} for details on the syntax of debug options.  This
     * method attempts no error checking of the options string; any exception
     * simply aborts the processing.
     * 
     * @param optionsMap A <code>java.util.HashMap</code> into which this method
     *                   will insert options settings.
     * @param options Input <code>String</code> of debug options.
     */
    private static void installOptions(HashMap optionsMap, String options) {
        ArrayList optionStack = new ArrayList();
        ArrayList valueStack = new ArrayList();
        StringTokenizer tokenizer = new StringTokenizer(options, ":()", true);
        boolean processedValue = true;
        Object currentOption = "";
        Object currentValue = "";
        int parenCount = 0;

        optionStack.add("");
        valueStack.add("");

        while (tokenizer.hasMoreTokens()) {
            String token = tokenizer.nextToken();

            if (":".equals(token)) {
                if (parenCount > 0) {
                    currentValue = currentValue + token;
                }
                if (!processedValue) {
                    optionsMap.put(currentOption, null);
                    currentOption = optionStack.get(optionStack.size()-1);
                }
            } else if ("(".equals(token)) {
                valueStack.add(currentValue + token);
                optionStack.add(currentOption);
                currentValue = "";
                parenCount++;
            } else if (")".equals(token)) {
                if (!processedValue) {
                    optionsMap.put(currentOption, null);
                }
                currentOption =
                    optionStack.remove(optionStack.size() - 1);
                optionsMap.put(currentOption, currentValue);
                currentOption = optionStack.get(optionStack.size() - 1);
                int newValueStackSize = valueStack.size()-1;
                currentValue = valueStack.remove(newValueStackSize).toString()
                                   + currentValue + token;
                valueStack.add(currentValue);
                parenCount--;
                processedValue = true;
            } else {
                if (parenCount > 0) {
                    currentValue = currentValue + token;
                }

                currentOption = (currentOption.equals("")) ? token
                                    : currentOption + "." + token;
                processedValue = false;
            }
        }

        if (!processedValue && !currentOption.equals("")) {
            optionsMap.put(currentOption, null);
        }
    }

    /**
     * Was this debug option specified?
     * @param option The option to check
     * @return <code>true</code> if and only if the option was specified
     */
    public static boolean wasSpecified(String option) {
        final boolean optionSeen =
            (sOptions != null) ? sOptions.containsKey(option) : false;
        if (sDebugMe) {
            System.err.println(HiddenOptions.class.getName()
                    + ":  Calling wasSpecified(\"" + option
                    + "\"); result == " + optionSeen);
        }
        return optionSeen;
    }

    /**
     * Get the value of the debug option as a Boolean.  This is equivalent to
     * <code>Boolean.valueOf(getStringValue(option))</code>, except that this
     * method will return <code>null</code> if
     * <code>getStringValue(option)</code> is a null reference.
     *  
     * @param option The option to check
     * @return <code>Boolean.TRUE</code> if the string value of the option
     *         compares equal to "true" without regard to case; otherwise,
     *         if the string value is not <code>null</code>, returns
     *         <code>Boolean.FALSE</code>; otherwise, <code>null</code>.
     */
    public static Boolean getBooleanValue(String option) {
        final String stringValue =
            (sOptions != null) ? (String) sOptions.get(option) : null;
        final Boolean optionValue =
            (stringValue != null) ? Boolean.valueOf(stringValue) : null;

        if (sDebugMe) {
            System.err.println(HiddenOptions.class.getName()
                    + ":  Calling getBooleanValue(\"" + option
                    + "\"); result ==  " + optionValue);
        }
        return optionValue;
    }

    /**
     * What was the value of this debug option, if any?
     * 
     * @param option The option to check
     * @return The string value of the option, or <code>null</code> if the
     *         option had no value.
     */
    public static String getStringValue(String option) {
        String optionValue = (sOptions != null) ? (String) sOptions.get(option)
                                                : null;
        if (sDebugMe) {
            System.err.println(HiddenOptions.class.getName()
                    + ":  Calling getStringValue(\"" + option
                    + "\"); result ==  \"" + optionValue+"\"");
        }

        return optionValue;
    }
    
    /**
     * What was the value of this debug option, if any?
     * 
     * @param option The option to check
     * @return The string value of the option, or <code>null</code> if the
     *         option had no value.
     */
    public static boolean optionValueIs(String option, String value) {
        String optionValue = getStringValue(option);
        
        boolean ret = value.equals(optionValue);
        if (sDebugMe) {
            System.err.println(HiddenOptions.class.getName()
                    + ":  Calling stringValueIs(\"" + option
                    + "\", \""+value+"\"); result ==  " + ret);
        }

        return ret;
    }

    /**
     * Get the value of the debug option as an <code>Integer</code>.  This is
     * equivalent to <code>Integer.valueOf(getStringValue(option))</code>,
     * except that this method will return <code>null</code> if
     * <code>getStringValue(option)</code> is a null reference or if the string
     * value cannot be parsed as an integer.
     *  
     * @param option The option to check
     * @return The value of the option converted to an <code>Integer</code>, or
     *         <code>null</code> if the option value is <code>null</code> or
     *         cannot be converted to an <code>int</code>.
     */
    public static Integer getIntegerValue(String option) {
        final String stringValue =
            (sOptions != null) ? (String) sOptions.get(option) : null;
        Integer optionValue = null;
        
        if (stringValue != null) {
            try {
                optionValue = Integer.valueOf(stringValue);
            } catch(Throwable t) {
                // Might throw a NumberFormatException; if so, the value is null
            }
        }

        if (sDebugMe) {
            System.err.println(HiddenOptions.class.getName()
                    + ":  Calling getIntegerValue(\"" + option
                    + "\"); result ==  " + optionValue);
        }

        return optionValue;
    }

    /**
     * Format debug option settings as a string.
     * @return The formatted debug options map
     */
    private static String dump() {
        return (sOptions != null) ? sOptions.toString() : "{}";
    }

    /*
    public static final void main(String[] args) {
        System.out.println(HiddenOptions.class.getName() + ":  " + dump());
        System.out.println("");
    }
    */
}
